@@ -17,6 +17,7 @@ import base64
 import errno
 import functools
 import importlib
+import multiprocessing
 import os
 import platform
 import queue
@@ -24,11 +25,12 @@ import re
 import subprocess
 import sys
 import weakref
+from collections import OrderedDict
 from configparser import ConfigParser, NoOptionError, NoSectionError
 from datetime import datetime
 from operator import itemgetter, methodcaller
 from statistics import mean
-from typing import Any, Dict, List, Union
+from typing import Any, Optional, Union
 from urllib.error import HTTPError, URLError
 from urllib.parse import urlparse
 from urllib.request import Request, urlopen
@@ -95,6 +97,11 @@ viewkeys = methodcaller('keys')
 viewvalues = methodcaller('values')
 viewitems = methodcaller('items')
 
+# Multiprocessing start method (on POSIX system)
+if LINUX or BSD or SUNOS or MACOS:
+    ctx_mp_fork = multiprocessing.get_context('fork')
+else:
+    ctx_mp_fork = multiprocessing.get_context()
 
 ###################
 # GLOBALS FUNCTIONS
@@ -126,18 +133,6 @@ def listvalues(d):
     return list(d.values())
 
 
-def iteritems(d):
-    return iter(d.items())
-
-
-def iterkeys(d):
-    return iter(d.keys())
-
-
-def itervalues(d):
-    return iter(d.values())
-
-
 def u(s, errors='replace'):
     if isinstance(s, text_type):
         return s
@@ -268,13 +263,45 @@ def safe_makedirs(path):
             raise
 
 
-def pretty_date(ref=False, now=datetime.now()):
+def get_time_diffs(ref, now):
+    if isinstance(ref, int):
+        diff = now - datetime.fromtimestamp(ref)
+    elif isinstance(ref, datetime):
+        diff = now - ref
+    elif not ref:
+        diff = 0
+
+    return diff
+
+
+def get_first_true_val(conds):
+    return next(key for key, val in conds.items() if val)
+
+
+def maybe_add_plural(count):
+    return "s" if count > 1 else ""
+
+
+def build_str_when_more_than_seven_days(day_diff, unit):
+    scale = {'week': 7, 'month': 30, 'year': 365}[unit]
+
+    count = day_diff // scale
+
+    return str(count) + " " + unit + maybe_add_plural(count)
+
+
+def pretty_date(ref, now=None):
     """
     Get a datetime object or a int() Epoch timestamp and return a
     pretty string like 'an hour ago', 'Yesterday', '3 months ago',
     'just now', etc
     Source: https://stackoverflow.com/questions/1551382/user-friendly-time-format-in-python
+
+    Refactoring done in commit https://github.com/nicolargo/glances/commit/f6279baacd4cf0b27ca10df6dc01f091ea86a40a
+    break the function. Get back to the old fashion way.
     """
+    if not now:
+        now = datetime.now()
     if isinstance(ref, int):
         diff = now - datetime.fromtimestamp(ref)
     elif isinstance(ref, datetime):
@@ -303,12 +330,15 @@ def pretty_date(ref=False, now=datetime.now()):
     if day_diff == 1:
         return "yesterday"
     if day_diff < 7:
-        return str(day_diff) + " days"
+        return str(day_diff) + " days" if day_diff > 1 else "a day"
     if day_diff < 31:
-        return str(day_diff // 7) + " weeks" if (day_diff // 7) > 1 else "1 week"
+        week = day_diff // 7
+        return str(week) + " weeks" if week > 1 else "a week"
     if day_diff < 365:
-        return str(day_diff // 30) + " months" if (day_diff // 30) > 1 else "1 month"
-    return str(day_diff // 365) + " years" if (day_diff // 365) > 1 else "1 year"
+        month = day_diff // 30
+        return str(month) + " months" if month > 1 else "a month"
+    year = day_diff // 365
+    return str(year) + " years" if year > 1 else "an year"
 
 
 def urlopen_auth(url, username, password):
@@ -334,11 +364,18 @@ def json_dumps(data) -> bytes:
     return b(res)
 
 
-def json_loads(data: Union[str, bytes, bytearray]) -> Union[Dict, List]:
+def json_loads(data: Union[str, bytes, bytearray]) -> Union[dict, list]:
     """Load a JSON buffer into memory as a Python object"""
     return json.loads(data)
 
 
+def list_to_dict(data):
+    """Convert a list of dict (with key in 'key') to a dict with key as key and value as value."""
+    if not isinstance(data, list):
+        return None
+    return {item[item['key']]: item for item in data if 'key' in item}
+
+
 def dictlist(data, item):
     if isinstance(data, dict):
         try:
@@ -357,13 +394,81 @@ def dictlist(data, item):
         return None
 
 
-def json_dumps_dictlist(data, item):
+def dictlist_json_dumps(data, item):
     dl = dictlist(data, item)
     if dl is None:
         return None
     return json_dumps(dl)
 
 
+def dictlist_first_key_value(data: list[dict], key, value) -> Optional[dict]:
+    """In a list of dict, return first item where key=value or none if not found."""
+    try:
+        ret = next(item for item in data if key in item and item[key] == value)
+    except StopIteration:
+        ret = None
+    return ret
+
+
+def auto_unit(number, low_precision=False, min_symbol='K', none_symbol='-'):
+    """Make a nice human-readable string out of number.
+
+    Number of decimal places increases as quantity approaches 1.
+    CASE: 613421788        RESULT:       585M low_precision:       585M
+    CASE: 5307033647       RESULT:      4.94G low_precision:       4.9G
+    CASE: 44968414685      RESULT:      41.9G low_precision:      41.9G
+    CASE: 838471403472     RESULT:       781G low_precision:       781G
+    CASE: 9683209690677    RESULT:      8.81T low_precision:       8.8T
+    CASE: 1073741824       RESULT:      1024M low_precision:      1024M
+    CASE: 1181116006       RESULT:      1.10G low_precision:       1.1G
+
+    :low_precision: returns less decimal places potentially (default is False)
+                    sacrificing precision for more readability.
+    :min_symbol: Do not approach if number < min_symbol (default is K)
+    :decimal_count: if set, force the number of decimal number (default is None)
+    """
+    if number is None:
+        return none_symbol
+    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
+    if min_symbol in symbols:
+        symbols = symbols[symbols.index(min_symbol) :]
+    prefix = {
+        'Y': 1208925819614629174706176,
+        'Z': 1180591620717411303424,
+        'E': 1152921504606846976,
+        'P': 1125899906842624,
+        'T': 1099511627776,
+        'G': 1073741824,
+        'M': 1048576,
+        'K': 1024,
+    }
+
+    if number == 0:
+        # Avoid 0.0
+        return '0'
+
+    # If a value is a float, decimal_precision is 2 else 0
+    decimal_precision = 2 if isinstance(number, float) else 0
+    for symbol in reversed(symbols):
+        value = float(number) / prefix[symbol]
+        if value > 1:
+            decimal_precision = 0
+            if value < 10:
+                decimal_precision = 2
+            elif value < 100:
+                decimal_precision = 1
+            if low_precision:
+                if symbol in 'MK':
+                    decimal_precision = 0
+                else:
+                    decimal_precision = min(1, decimal_precision)
+            elif symbol in 'K':
+                decimal_precision = 0
+            return '{:.{decimal}f}{symbol}'.format(value, decimal=decimal_precision, symbol=symbol)
+
+    return f'{number:.{decimal_precision}f}'
+
+
 def string_value_to_float(s):
     """Convert a string with a value and an unit to a float.
     Example:
@@ -431,18 +536,32 @@ def folder_size(path, errno=0):
         return ret_size, ret_err
 
 
-def weak_lru_cache(maxsize=128, typed=False):
+def _get_ttl_hash(ttl):
+    """A simple (dummy) function to return a hash based on the current second.
+    TODO: Implement a real TTL mechanism.
+    """
+    if ttl is None:
+        return 0
+    now = datetime.now()
+    return now.second
+
+
+def weak_lru_cache(maxsize=1, typed=False, ttl=None):
     """LRU Cache decorator that keeps a weak reference to self
+
+    Warning: When used in a class, the class should implement __eq__(self, other) and __hash__(self) methods
+
     Source: https://stackoverflow.com/a/55990799"""
 
     def wrapper(func):
         @functools.lru_cache(maxsize, typed)
-        def _func(_self, *args, **kwargs):
+        def _func(_self, *args, ttl_hash=None, **kwargs):
+            del ttl_hash  # Unused parameter, but kept for compatibility
             return func(_self(), *args, **kwargs)
 
         @functools.wraps(func)
         def inner(self, *args, **kwargs):
-            return _func(weakref.ref(self), *args, **kwargs)
+            return _func(weakref.ref(self), *args, ttl_hash=_get_ttl_hash(ttl), **kwargs)
 
         return inner
 
@@ -451,7 +570,7 @@ def weak_lru_cache(maxsize=128, typed=False):
 
 def namedtuple_to_dict(data):
     """Convert a namedtuple to a dict, using the _asdict() method embedded in PsUtil stats."""
-    return {k: (v._asdict() if hasattr(v, '_asdict') else v) for k, v in data.items()}
+    return {k: (v._asdict() if hasattr(v, '_asdict') else v) for k, v in list(data.items())}
 
 
 def list_of_namedtuple_to_list_of_dict(data):
@@ -463,3 +582,52 @@ def replace_special_chars(input_string, by=' '):
     """Replace some special char by another in the input_string
     Return: the string with the chars replaced"""
     return input_string.replace('\r\n', by).replace('\n', by).replace('\t', by)
+
+
+def atoi(text):
+    return int(text) if text.isdigit() else text
+
+
+def natural_keys(text):
+    """Return a text in a natural/human readable format."""
+    return [atoi(c) for c in re.split(r'(\d+)', text)]
+
+
+def exit_after(seconds, default=None):
+    """Exit the function if it takes more than 'seconds' seconds to complete.
+    In this case, return the value of 'default' (default: None)."""
+
+    def handler(q, func, args, kwargs):
+        q.put(func(*args, **kwargs))
+
+    def decorator(func):
+        if not LINUX:
+            return func
+
+        def wraps(*args, **kwargs):
+            try:
+                q = ctx_mp_fork.Queue()
+            except PermissionError:
+                # Manage an exception in Snap packages on Linux
+                # The strict mode prevent the use of multiprocessing.Queue()
+                # There is a "dirty" hack:
+                # https://forum.snapcraft.io/t/python-multiprocessing-permission-denied-in-strictly-confined-snap/15518/2
+                # But i prefer to just disable the timeout feature in this case
+                func(*args, **kwargs)
+            else:
+                p = ctx_mp_fork.Process(target=handler, args=(q, func, args, kwargs))
+                p.start()
+                p.join(timeout=seconds)
+                if not p.is_alive():
+                    return q.get()
+                p.terminate()
+                p.join(timeout=0.1)
+                if p.is_alive():
+                    # Kill in case processes doesn't terminate
+                    # Happens with cases like broken NFS connections
+                    p.kill()
+            return default
+
+        return wraps
+
+    return decorator
